@@ -7,10 +7,12 @@ import inspect
 from zcross.raw.binding import *
 import zcross.raw.binding as raw_binding
 
+from collections import OrderedDict
+
 import pyxb.utils.domutils
 pyxb.utils.domutils.BindingDOMSupport.SetDefaultNamespace("https://zcross.net")
 
-re_reference   = re.compile(r'@(\w+){([\w\-]+),([\s\S]*)}', re.MULTILINE)
+re_reference   = re.compile(r'\s*@(\w+){([\w\-]+),([\s\S]*)}', re.MULTILINE)
 re_whitespace  = re.compile(r'\s+', re.MULTILINE)
 re_database_id = re.compile(r'\<database\s+id\s*=\s*\"([A-Za-z0-9]+)\"', re.MULTILINE)
 
@@ -23,45 +25,57 @@ def save_to_xml(document, filename):
     with open(filename, 'w') as f:
         f.write(document.toxml("utf-8", element_name='zcross').decode('utf-8'))
 
-def _get_data_dir():
-    data_dir = None
+def _get_data_dirs():
+    data_dirs = []
     
-    if 'ZCROSS_DATA' in os.environ and os.path.exists(os.environ['ZCROSS_DATA']):
-        data_dir = os.environ['ZCROSS_DATA']  
-    elif os.path.exists('/opt/zcross_data'):
-        data_dir = '/opt/zcross_data'
+    if 'ZCROSS_DATA' in os.environ:
+        for token in os.environ['ZCROSS_DATA'].split(':'):
+            if token:
+                if os.path.exists(token) and os.path.isdir(token):
+                    data_dirs.append(token)
+                else:
+                    raise Exception('Unable to access the directory \'{}\' specified in path ZCROSS_DATA env.'.format(token))  
+                    
+    elif os.path.exists('/opt/zcross_data') and os.path.isdir('/opt/zcross_data'):
+        data_dirs.append('/opt/zcross_data')
         
-    if data_dir is None:
-        raise Exception('Unable to find ZCross data file: define ZCROSS_DATA env variable.')
+    if data_dirs is None:
+        raise Exception('Unable to find ZCross database directory path: define ZCROSS_DATA env variable.')
       
-    return data_dir
+    return data_dirs
+    
     
 def load_by_name(database):
   
-    for r, d, f in os.walk(_get_data_dir()):
-        for fx in f:
-            if fx.endswith('.xml'):
-                filename = os.path.join(r, fx)
-                with open(filename, 'r') as f:
-                    header = f.read(1024) # reading the first caracters of the XML files whitout loading everything in memory
-                    m = re_database_id.search(header)
-                    if m and m.group(1).lower() == database.lower():
-                        f.seek(0)
-                        return raw_binding.CreateFromDocument(f.read())
+    for directory in _get_data_dirs():
+        for r, d, f in os.walk(directory):
+            for fx in f:
+                if fx.endswith('.xml'):
+                    filename = os.path.join(r, fx)
+                    with open(filename, 'r') as f:
+                        header = f.read(1024) # reading the first caracters of the XML files whitout loading everything in memory
+                        m = re_database_id.search(header)
+                        if m and m.group(1).lower() == database.lower():
+                            f.seek(0)
+                            return raw_binding.CreateFromDocument(f.read())                    
+    raise Exception('Unable to find a database with name \'{}\''.format(database))
+                        
+     
     
 def load_all(verbose = False):
   
     contents = []
    
-    for r, d, f in os.walk(_get_data_dir()):
-        for fx in f:
-            if fx.endswith('.xml'):
-                filename = os.path.join(r, fx)
+    for directory in _get_data_dirs():
+        for r, d, f in os.walk(directory):
+            for fx in f:
+                if fx.endswith('.xml'):
+                    filename = os.path.join(r, fx)
 
-                with open(filename, 'r') as f:
-                    if verbose:
-                        print('Loading {}'.format(filename))
-                    contents.append(raw_binding.CreateFromDocument(f.read()))
+                    with open(filename, 'r') as f:
+                        if verbose:
+                            print('Loading {}'.format(filename))
+                        contents.append(raw_binding.CreateFromDocument(f.read()))
     return contents
                     
 
@@ -69,11 +83,88 @@ def load_all(verbose = False):
 class ZCross(raw_binding.ZCross):
     pass
     
+class DatabaseIterator:
+
+    def __init__(self, database):
+        self._child_iter = iter(database.groups.orderedContent())
+        
+    def __next__(self):
+       return next(self._child_iter).value
+    
+class Database(raw_binding.Database):
+    
+    def __str__(self):
+        return "Database({})".format(self.id)
+        
+
+    def __iter__(self):
+        return DatabaseIterator(self)
+        
+    def __len__(self):
+        return len(self.groups.orderedContent())
+    
+    def __getitem__(self, key):
+        content = self.groups.orderedContent()
+        if isinstance(key, int):
+            return content[key].value
+        else:
+            return next(c.value for c in content if c.value.id == key)
+        
+class ReferencesIterator:
+
+    def __init__(self, references):
+        self._child_iter = iter(references.reference)
+        
+    def __next__(self):
+       return next(self._child_iter)
+
+class References(raw_binding.References):
+        
+    def __iter__(self):
+        return ReferencesIterator(self)
+        
+    def __len__(self):
+        return len(self.reference.orderedContent())
+    
+    def __getitem__(self, key):
+        content = self.reference.orderedContent()
+        if isinstance(key, int):
+            return content[key].value
+        else:
+            return next(c.value for c in content if c.value.label == key)
+    
+        
+class GroupIterator:
+
+    def __init__(self, group):
+        self._child_iter = iter(group.processes.orderedContent())
+        
+    def __next__(self):
+       return next(self._child_iter).value
+       
+class Group(raw_binding.Group):
+
+    def __str__(self):
+        return "Group({})".format(self.id)
+        
+    def __iter__(self):
+        return GroupIterator(self)
+    
+    def __len__(self):
+        return len(self.processes.orderedContent())
+    
+    def __getitem__(self, key):
+        content = self.processes.orderedContent()
+        if isinstance(key, int):
+            return content[key].value
+    
 class Reference(raw_binding.Reference):
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
          
+        self._payload = {}
+        
         m = re_reference.match(self.value())
         
         if m:
@@ -119,12 +210,22 @@ class Reference(raw_binding.Reference):
                         value = value.strip('{').strip('}')
                         
                     setattr(self, key, value)
+                    self._payload[key] = value
                     
                     current_key   = ''
                     current_value = ''
                     level = 0
                     section = 0
-       
+        else:
+            raise ValueError('Unable to parse reference:\n{}\n'.format(self.value()))
+    
+    def payload(self):
+        return self._payload
+        
+    def items(self):
+        for k, v in self._payload.items():
+            yield k,v
+        
     def bibtex(self):
         s = '@{}{{{},\n'.format(self.type, self.label)
         for k,v in vars(self).items():
@@ -134,17 +235,115 @@ class Reference(raw_binding.Reference):
         
         return s
         
+        
+class ProcessIterator:
+
+    def __init__(self, process):
+        self._child_iter = iter(process._data.items())
+        
+    def __next__(self):
+       return next(self._child_iter)
+       
       
 class Process(raw_binding.Process):
     
-    def __str__(self):
-        s = ''      
-        s += ' + '.join(str(r.value) for r in self.reactants.orderedContent())
-        s += ' → '
-        s += ' + '.join(str(p.value) for p in self.products.orderedContent())
-        return s
+    def _initialize(self):
         
-  
+        if not hasattr(self, '_data'):
+            self._data = OrderedDict(zip(self.data_x.value(), self.data_y.value()))
+
+    def __iter__(self):
+        self._initialize()
+        return ProcessIterator(self)
+    
+    def __len__(self):
+        self._initialize()
+        return len(self._data)
+    
+    def __getitem__(self, key):
+        self._initialize()
+        if isinstance(key, int):
+            return list(self._data.items())[key]
+
+    def get_data(self):
+        self._initialize()
+        return self._data
+    
+            
+    def get_reaction(self):
+        s  = ' + '.join(str(r.value) for r in self.reactants.orderedContent())
+        s += ' → '
+        # This hurts :)
+        s += ' or '.join(' + '.join(str(p.value) for p in ps.orderedContent()) for ps in self.products)
+        return s
+    
+    @property
+    def energy_units(self):
+        return self.data_x.units
+        
+    @property
+    def cross_section_units(self):
+        return self.data_y.units
+        
+    def __str__(self):
+        return 'Process({}: {})'.format(self.id, self.get_reaction())      
+        
+    def get_nano_type(self):
+        if self.collisionType == 'elastic':
+            if self.momentOrder == 0:
+                return  'el'
+            elif self.momentOrder == 1:
+                return  'mt'
+            elif self.momentOrder == 2:
+                return  'vs'
+            else:
+                return  'mo {}'.format(self.moment_order)
+        elif self.collisionType == 'inelastic':
+            if   self.inelasticType == 'excitation_ele':
+                return  'ele'
+            elif self.inelasticType == 'excitation_vib':
+                return  'vib'
+            elif self.inelasticType == 'excitation_rot':
+                return  'rot'
+            elif self.inelasticType == 'ionization':
+                return  'ion'
+            elif self.inelasticType == 'attachment':
+                return  'att'
+            elif self.inelasticType == 'neutral':
+                return  'neu'
+            else:
+                return  '???'
+        else:
+            return  '???'   
+        
+    def get_tiny_type(self):
+        if self.collisionType == 'elastic':
+            if self.momentOrder == 0:
+                return  'ela el'
+            elif self.momentOrder == 1:
+                return  'ela mt'
+            elif self.momentOrder == 2:
+                return  'ela vs'
+            else:
+                return  'ela {}'.format(self.moment_order)
+        elif self.collisionType == 'inelastic':
+            if   self.inelasticType == 'excitation_ele':
+                return  'exc ele'
+            elif self.inelasticType == 'excitation_vib':
+                return  'exc vib'
+            elif self.inelasticType == 'excitation_rot':
+                return  'exc rot'
+            elif self.inelasticType == 'ionization':
+                return  'ion'
+            elif self.inelasticType == 'attachment':
+                return  'att'
+            elif self.inelasticType == 'neutral':
+                return  'neu'
+            else:
+                return  '???'
+        else:
+            return  '???'
+              
     def get_simple_type(self):
         if self.collisionType == 'elastic':
             if self.momentOrder == 0:
@@ -172,6 +371,14 @@ class Process(raw_binding.Process):
                 return  '???'
         else:
             return  '???'
+            
+        
+    def payload(self):
+        return self._payload
+        
+    def items(self):
+        for k, v in self._payload.items():
+            yield k,v
    
         
 class Electron(raw_binding.Electron):
@@ -204,6 +411,25 @@ class Molecule(raw_binding.Molecule):
         
         return s
         
+    def get_latex(self):
+        
+        s = self.orderedContent()[0].value
+        
+        if self.charge != None:
+            if abs(self.charge) > 1:
+                s += '^{' + str(abs(self.charge)) + ('+' if self.charge > 0 else '-')  + '}'
+                
+            elif self.charge > 0:
+                s += '^{' + '+' * self.charge + '}'
+                
+            elif self.charge < 0:
+                 s += '^{' + '-' * (-self.charge) + '}'
+                
+        if self.state != None:
+            s += '\text{(' + self.state + ')}'
+        
+        return s
+        
         
     # def explode(self):
         
@@ -276,7 +502,7 @@ class Molecule(raw_binding.Molecule):
     
 class Parameter(raw_binding.Parameter):
     def __str__(self):
-        return self.name + ' = ' + self.value + (' ' + self.units if hasattr(self, 'units') else '')
+        return self.name + ' = ' + str(self.value()) + (' ' + self.units if self.units is not None else '')
     
    
     
@@ -319,10 +545,13 @@ def get_scripted_number(number, superscript, forcePlus):
 
 
 raw_binding.ZCross._SetSupersedingClass(ZCross)
+raw_binding.Database._SetSupersedingClass(Database)
+raw_binding.Group._SetSupersedingClass(Group)
 raw_binding.Electron._SetSupersedingClass(Electron)
 raw_binding.Molecule._SetSupersedingClass(Molecule)
 raw_binding.Parameter._SetSupersedingClass(Parameter)
 raw_binding.Process._SetSupersedingClass(Process)
+raw_binding.References._SetSupersedingClass(References)
 raw_binding.Reference._SetSupersedingClass(Reference)
 
 
